import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;
import ij.plugin.frame.*;
import ij.io.SaveDialog;
import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.event.*;
import javax.swing.SwingConstants;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.table.DefaultTableModel;
import java.io.*;
import java.util.ArrayList;
import java.util.Vector;
import org.apache.sanselan.*;
import org.apache.sanselan.common.*;
import org.apache.sanselan.formats.jpeg.*;
import org.apache.sanselan.formats.tiff.*;
import org.apache.sanselan.formats.tiff.constants.*;
import org.apache.sanselan.formats.tiff.write.TiffOutputSet;
import org.apache.sanselan.formats.tiff.write.TiffOutputField;
import org.apache.sanselan.formats.tiff.write.TiffOutputDirectory;
import org.apache.sanselan.formats.jpeg.exifRewrite.ExifRewriter;

public class Exif_Edit implements PlugIn, ActionListener {
	private ImagePlus img;
	//thumbnail
	private int offset;
	private byte[] thumb;
	private BufferedImage originalThumbnail = null;
	private BufferedImage thumbnail = null;
	
	private TiffImageMetadata.GPSInfo gps;
	
	private JLabel labelThumbnail;
	private ExifModel model;
	private File input = null;
	private File newImage = null;
	private ArrayList<TagInfo> tag = null;
	private ArrayList listExif = null;
	
	private boolean thumbChange = false;
	private boolean delThumb = false; 
	private boolean removeExif = false;
	private boolean repExif = false;
	
	private JButton update;
	private JButton replace;
	private JButton delete;
	private JButton insert;
	private JButton replaceExif;
	
	public void run(String arg) {
		JFileChooser chooser = new JFileChooser();
		FileNameExtensionFilter filter = new FileNameExtensionFilter("Images", "jpg");
		chooser.setFileFilter(filter);
		int returnVal = chooser.showOpenDialog(null);
		if(returnVal == JFileChooser.APPROVE_OPTION) {
			input=chooser.getSelectedFile();
			img = new ImagePlus(input.getPath());
			img.show();
			readExif(input);
		}
		else {
			return;
		}
		
		//frame
		JFrame window = new JFrame(img.getTitle() + "     Exif Edit - By Coppoletta Claudio & Russo Placido");
		window.setResizable(false);
		window.setBounds(150, 150, 900, 550);
		window.setLayout(new FlowLayout(FlowLayout.CENTER));
		
		JPanel up = new JPanel();
		//thumbnail
		JPanel panelThumbnail = new JPanel(new GridLayout(2,1));
		labelThumbnail = new JLabel();
		labelThumbnail.setHorizontalAlignment(SwingConstants.CENTER);
		if (extractThumbnail(input))
			labelThumbnail.setIcon(new ImageIcon(thumbnail));
		else
			labelThumbnail.setText("Thumbnail not found!");
		panelThumbnail.add(labelThumbnail);
		JPanel thumbButton = new JPanel();
		update = new JButton("Update");
		update.addActionListener(this);
		thumbButton.add(update);
		replace = new JButton("Replace");
		replace.addActionListener(this);
		thumbButton.add(replace);
		delete = new JButton("Delete");
		delete.addActionListener(this);
		thumbButton.add(delete);
		panelThumbnail.add(thumbButton);
		//Exif
		JPanel panelExif = new JPanel();
		model = new ExifModel();
		JTable table = new JTable(model);
		table.getTableHeader().getColumnModel().getColumn(0).setMaxWidth(200);
		table.getTableHeader().getColumnModel().getColumn(1).setMaxWidth(300);
		table.getTableHeader().getColumnModel().getColumn(2).setMaxWidth(50);
		table.getTableHeader().getColumnModel().getColumn(3).setMaxWidth(50);
		showTable();
		JScrollPane scrollPane = new JScrollPane();
		table.setSurrendersFocusOnKeystroke(true);
		table.setBackground(Color.WHITE);
		scrollPane.setViewportView(table);
		
		up.add(panelThumbnail);
		up.add(scrollPane);
		window.add(up);
		
		//bottoni
		JPanel down = new JPanel();
		JButton save = new JButton("Save");
		save.addActionListener(this);
		down.add(save);
		insert = new JButton("Insert");
		insert.addActionListener(this);
		down.add(insert);
		replaceExif = new JButton("Replace Exif");
		replaceExif.addActionListener(this);
		down.add(replaceExif);
		JButton deleteAll = new JButton("Delete All");
		deleteAll.addActionListener(this);
		down.add(deleteAll);
		JButton reset = new JButton("Reset");
		reset.addActionListener(this);
		down.add(reset);
		
		window.add(down);
		window.setVisible(true);
	}	
	
	private void showTable() {
		Object[] row = new Object[4];
		
		while (model.getRowCount() != 0)
			model.removeRow(0);
		
		TiffImageMetadata.Item item;
		int size = 0;
		if (listExif != null) size=listExif.size();
		
		for (int i=0; i<size; i++) {
			item = (TiffImageMetadata.Item) listExif.get(i);
			try {
				row[0] = item;
				row[1] = item.getTiffField().getValueDescription();
			}
			catch (Exception e) {
				IJ.showMessage("ERROR: "+e.getMessage());
			}
			row[2] = new Boolean(false);
			row[3] = new Boolean(false);
			model.addRow(row);
		}
		//GPS
		if (gps != null) {
			try {
				double longitude = gps.getLongitudeAsDegreesEast();
				double latitude = gps.getLatitudeAsDegreesNorth();
				row[0] = "GPS Description";
				row[1] = gps;
				row[2] = new Boolean(false);
				row[3] = new Boolean(false);
				model.addRow(row);
				row[0] = "GPS Latitude (Degrees North)";
				row[1] = latitude;
				row[2] = new Boolean(false);
				row[3] = new Boolean(false);
				model.addRow(row);
				row[0] = "GPS Longitude (Degrees East)";
				row[1] = longitude;
				row[2] = new Boolean(false);
				row[3] = new Boolean(false);
				model.addRow(row);
			}
			catch (Exception e) {
				IJ.showMessage("ERROR: "+e.getMessage());
			}
		}
	}
	
	private void readExif(File file) {
		IImageMetadata metadata = null;
		TiffImageMetadata exif = null;
		try {
			metadata = Sanselan.getMetadata(file);
		}
		catch (Exception e) {
			IJ.showMessage("ERROR: "+e.getMessage());
		}
		
		if (metadata instanceof JpegImageMetadata) {
			JpegImageMetadata jpeg = (JpegImageMetadata) metadata;
			exif = jpeg.getExif();
			try {
				listExif = exif.getItems();
				
				//GPS
				if (exif != null)
					gps = exif.getGPS();
			}
			catch (Exception e) {
				IJ.showMessage("ERROR: "+e.getMessage());
			}
		}
		else {
			listExif = null;
			gps = null;
		}
		
		createListTag();
	}
	
	private boolean extractThumbnail(File file) {
		IImageMetadata metadata = null;
		thumb = null;
		
		try {
			metadata = Sanselan.getMetadata(file);
		}
		catch (Exception e) {
			IJ.showMessage("ERROR: "+e.getMessage());
		}
		
		if (metadata instanceof JpegImageMetadata) {
			JpegImageMetadata jpeg = (JpegImageMetadata) metadata;
			JpegImageData thumbImage;
			ArrayList dirs = jpeg.getExif().getDirectories();
			for (int i=0; i<dirs.size(); i++) {
				TiffImageMetadata.Directory tiffdir = (TiffImageMetadata.Directory) dirs.get(i);
				if(tiffdir.getJpegImageData() != null){
					thumbImage = tiffdir.getJpegImageData();
					thumb = thumbImage.data;
					if(thumb != null) {
						try {
							InputStream in = new ByteArrayInputStream(thumb);
							originalThumbnail = thumbnail = ImageIO.read(in);
							offset = thumbImage.offset; //offset
							return true;
						}
						catch (Exception e) {
							IJ.showMessage("ERROR: "+e.getMessage());
						}
					}
				}
			}
		}
		return false;
	}
	
	public void createListTag() {
		TagInfo[] tiffTag = TiffTagConstants.ALL_TIFF_TAGS;
		TagInfo[] exifTag = ExifTagConstants.ALL_EXIF_TAGS;
		tag = new ArrayList<TagInfo>();
		TiffImageMetadata.Item item;
		TagInfo t;
		int size = 0;
		if (listExif != null) size=listExif.size();
		boolean present;
		TagInfo x;
		int l = tiffTag.length;
		int sizeTag = l + exifTag.length;
		for (int i=0; i<sizeTag; i++) {
			present = false;
			if (i<l)
				x = tiffTag[i];
			else
				x = exifTag[i-l];
			if (x.isOffset() || x.isUnknown())
				continue;
			for (int j=0; j<size; j++) {
				item = (TiffImageMetadata.Item) listExif.get(j);
				t = item.getTiffField().tagInfo;
				if (t.tag == x.tag) {
					present=true;
					break;
				}
			}
			if (!present)
				tag.add(x);
		}
	}
	
	public void setThumbnail(BufferedImage image) {
		int x = (160 * image.getHeight())/image.getWidth();
		BufferedImage newThumb = new BufferedImage(160,x,BufferedImage.TYPE_INT_RGB);
		//rescale image
		Graphics2D g = newThumb.createGraphics();
		g.drawImage(image, 0, 0, 160, x, null);
		g.dispose();
		thumbnail = newThumb;
		//show Image
		labelThumbnail.setText(null);
		labelThumbnail.setIcon(new ImageIcon(thumbnail));
		thumbChange = true;
	}
		
	//per i bottoni
	public void actionPerformed(ActionEvent e) {
		if(e.getActionCommand().equals("Save")) {
			
			SaveDialog fileSave= new SaveDialog("Save file", "new"+img.getTitle(), ".jpg");
			String name = fileSave.getFileName();
			String dir=fileSave.getDirectory();
			if (name != null) {
				if (input.getPath().equals(dir+name))
					name = name.substring(0,name.length()-4)+" (1).jpg";
				save(input, dir+name);
			}
		}
		if(e.getActionCommand().equals("Reset")) {
			//Thumbnail
			if (originalThumbnail != null)
				labelThumbnail.setIcon(new ImageIcon(originalThumbnail));
			else
				labelThumbnail.setIcon(null);
			thumbChange = false;
			removeExif = false;
			delThumb = false;
			repExif = false;
			update.setEnabled(true);
			replace.setEnabled(true);
			delete.setEnabled(true);
			insert.setEnabled(true);
			replaceExif.setEnabled(true);
			readExif(input);
			showTable();
			IJ.showMessage("Reset");
		}
		if(e.getActionCommand().equals("Insert")) {
			insert();
		}
		if(e.getActionCommand().equals("Update")) {
			setThumbnail(img.getBufferedImage());
			delThumb=false;
			IJ.showMessage("Update thumbnail");
		}
		if(e.getActionCommand().equals("Replace")) {
			File file = null;
			ImagePlus newImg = null;
			JFileChooser chooser = new JFileChooser();
			FileNameExtensionFilter filter = new FileNameExtensionFilter("Images", "jpg");
			chooser.setFileFilter(filter);
			int returnVal = chooser.showOpenDialog(null);
			if(returnVal == JFileChooser.APPROVE_OPTION) {
				file=chooser.getSelectedFile();
				newImg = new ImagePlus(file.getPath());
				newImg.show();
			}
			setThumbnail(newImg.getBufferedImage());
			delThumb=false;
			IJ.showMessage("Replace thumbnail");
		}
		if(e.getActionCommand().equals("Replace Exif")){
			JFileChooser chooser = new JFileChooser();
			FileNameExtensionFilter filter = new FileNameExtensionFilter("Images", "jpg");
			chooser.setFileFilter(filter);
			int returnVal = chooser.showOpenDialog(null);
			if(returnVal == JFileChooser.APPROVE_OPTION) {
				newImage=chooser.getSelectedFile();
				readExif(newImage);
				showTable();
				repExif = true;
				thumbChange = true;
				IJ.showMessage("Replace Exif metadata (image "+newImage.getName()+")");
			}
			else {
				return;
			}
		}
		if(e.getActionCommand().equals("Delete All")){
			removeExif=true;									
			int size = model.getRowCount();
			while (model.getRowCount() != 0)
				model.removeRow(0);
			labelThumbnail.setIcon(null);
			update.setEnabled(false);
			replace.setEnabled(false);
			delete.setEnabled(false);
			insert.setEnabled(false);
			replaceExif.setEnabled(false);
			IJ.showMessage("Delete all Exif data");
		}
		//cancella il thumbnail
		if(e.getActionCommand().equals("Delete")){
			delThumb = true;
			thumbChange = true;
			labelThumbnail.setIcon(null);
			IJ.showMessage("Delete thumbnail");
		}
	}
	
	public void save (final File in, String out) {
		OutputStream os = null;
		TiffImageMetadata exif = null;
		String result = "Image "+out+"\n\n";
		try {
			boolean noMetadata = false;
			TiffOutputSet outputSet = null;
			IImageMetadata metadata = null;
			if (repExif) {
				metadata = Sanselan.getMetadata(newImage);
				result = result + "Replace metadata Exif (image "+newImage.getName()+")\n";
			}
			else
				metadata = Sanselan.getMetadata(input);
			JpegImageMetadata jpeg = null;
			if (metadata instanceof JpegImageMetadata)
				jpeg = (JpegImageMetadata) metadata;
			else {
				noMetadata = true;
				jpeg = new JpegImageMetadata(null, new TiffImageMetadata(new TiffContents(new TiffHeader(TiffConstants.DEFAULT_TIFF_BYTE_ORDER, 42, 0), new ArrayList())));
			}
			if (jpeg != null) {
				exif = jpeg.getExif();
				if (exif != null) {
					outputSet = exif.getOutputSet();
				}
			}
			if (outputSet == null)
				outputSet = new TiffOutputSet();
			if (noMetadata)
				outputSet.addRootDirectory();
			
			if (!removeExif){
				TiffImageMetadata.Item item;
				Object o = null;
				int type = 0;
				String value = "";
				TiffOutputField field = null;
				TagInfo t = null;
				int size = 0;
				if (listExif != null) size = listExif.size();
				int sizeModel = model.getRowCount();
				for (int i=0; i<sizeModel; i++) {
					if ((Boolean) model.getValueAt(i,2)) {
						try {
							o = model.getItem(i);
							//x modifica tag esistenti
							if (o instanceof TiffImageMetadata.Item) {
								item = (TiffImageMetadata.Item) o;
								t = item.getTiffField().tagInfo;
								type=item.getTiffField().type;
							}
							//x metadati gps
							else if (o instanceof String) {
								double latitude = Double.parseDouble(model.getValueAt(++i,1).toString());
								double longitude = Double.parseDouble(model.getValueAt(++i,1).toString());
								outputSet.setGPSInDegrees(longitude, latitude);
								result = result + "Edit Tag GPS: latitude= "+latitude+" longitude= "+longitude+"\n";
								continue;
							}
							//x nuovi inserimenti
							else {
								t = (TagInfo) o;
								if (t.dataTypes != null)
									type = t.dataTypes[0].type;
							}
							
							if (t.isText() || type == 2) {	//String
								value = model.getValueAt(i,1).toString();
								if (value.length()>0 && value.charAt(0) == '\'') value=value.substring(1);
								if (value.length()>0 && value.charAt(value.length()-1) == '\'') value=value.substring(0,value.length()-1);
								field = new TiffOutputField (t, TiffConstants.FIELD_TYPE_ASCII, value.length(), value.getBytes());
								result = result + "Edit "+model.getValueAt(i,0)+": "+value+" (Type String)\n";
							}
							else if (t.length != 1 || type == 7) {		//byte[]
								Vector<Byte> r = new Vector<Byte>();
								value = model.getValueAt(i,1).toString()+" ";
								int count=0;
								int start=0;
								for (int j=0; j<value.length(); j++) {
									if (value.charAt(j) == ' ' || value.charAt(j) == ',') {
										if (start<j)
											r.add(Byte.parseByte(value.substring(start,j)));
										start=j+1;
									}
								}
								byte[] val = new byte[r.size()];
								for (int j=0; j<r.size(); j++)
									val[j] = (byte) r.get(j);
								
								field = new TiffOutputField(t, TiffConstants.FIELD_TYPE_BYTE, val.length, val);
								result = result + "Edit "+model.getValueAt(i,0)+": "+value+" (Type byte[])\n";
							}
							else if (type == 4 || type == 9) {	//Long
								value = model.getValueAt(i,1).toString();
								field = TiffOutputField.create(t, outputSet.byteOrder, new Integer(value));
								result = result + "Edit "+model.getValueAt(i,0)+": "+value+" (Type Long)\n";
							}
							else if (type == 3 || type == 8) {	//Short
								value = model.getValueAt(i,1).toString();
								field = TiffOutputField.create(t, outputSet.byteOrder, new Integer(value));
								result = result + "Edit "+model.getValueAt(i,0)+": "+value+" (Type Short)\n";
							}
							else if (type == 1 || type == 6) {	//Byte
								value = model.getValueAt(i,1).toString();
								field = TiffOutputField.create(t, outputSet.byteOrder, new Byte(value));
								result = result + "Edit "+model.getValueAt(i,0)+": "+value+" (Type Byte)\n";
							}
							else if (type == 12) {			//double
								value = model.getValueAt(i,1).toString();
								field = TiffOutputField.create(t, outputSet.byteOrder, new Double(value));
								result = result + "Edit "+model.getValueAt(i,0)+": "+value+" (Type Double)\n";
							}
							else if (type == 11) {			//float
								value = model.getValueAt(i,1).toString();
								field = TiffOutputField.create(t, outputSet.byteOrder, new Float(value));
								result = result + "Edit "+model.getValueAt(i,0)+": "+value+" (Type Float)\n";
							}
							else if (type == 5 || type == 10) {	//RationalNumber
								value = model.getValueAt(i,1).toString();
								int j=0;
								while (j<value.length() && (value.charAt(j) != '/' && value.charAt(j) != ',' && value.charAt(j) != '.')) j++;
								if (j==value.length())
									field = TiffOutputField.create(t,outputSet.byteOrder, new RationalNumber(Integer.parseInt(value),1));
								else {
									int start=j;
									while (start>0 && (value.charAt(start) != '(' && value.charAt(start) != ' ')) start--;
									int end=j;
									while (end<value.length() && (value.charAt(end) != ')' && value.charAt(end) != ' ')) end++;
									String num = value.substring(start,j);
									String den = value.substring(j+1,end);
									int n,d;
									if (value.charAt(j) == '/') {
										n = Integer.parseInt(num);
										d = Integer.parseInt(den);
									}
									else {
										n = Integer.parseInt(num+den);
										d = 1;
										for (int k=0; k<den.length(); k++) d *= 10;
									}
									field = TiffOutputField.create(t, outputSet.byteOrder, new RationalNumber(n, d));
								}
								result = result + "Edit "+model.getValueAt(i,0)+": "+value+" (Type RationalNumber)\n";
							}
							else {
								value = model.getValueAt(i,1).toString();
								result = result + "Can not edit tag "+model.getValueAt(i,0)+": "+value+" (Type Unknown)\n";
								continue;
							}
							
							TiffOutputDirectory dir = null;
							if (t.directoryType != null) {
								dir = outputSet.findDirectory(t.directoryType.directoryType);
								if (dir == null) {
									dir = new TiffOutputDirectory(t.directoryType.directoryType);
									outputSet.addDirectory(dir);
								}
								else
									dir.removeField(t);
								dir.add(field);
							}
							else {
								dir = new TiffOutputDirectory(0);	//nella root
								dir.add(field);
							}
						}
						catch (Exception e) {
							result = result + "Can not edit tag "+t.name+". ERROR: "+e.getMessage()+"\n";
						}
					}
					
					//Delete
					else if ((Boolean) model.getValueAt(i,3)) {
						try {
							o = model.getItem(i);
							//tag presenti
							if (o instanceof TiffImageMetadata.Item) {
								item = (TiffImageMetadata.Item) o;
								t = item.getTiffField().tagInfo;
								if (t.directoryType != null) {
									TiffOutputDirectory dir = outputSet.findDirectory(t.directoryType.directoryType);
									if (dir == null)
										result = result + "Can not remove tag "+t.name+". Directory: "+t.directoryType.name+" not found.\n";
									else {
										dir.removeField(t);
										result = result + "Remove tag "+t.name+"\n";
									}
								}
								else {
									if (outputSet.findField(t) == null)
										result = result + "Can not remove tag "+t.name+"\n";
									else {
										outputSet.removeField(t);
										result = result + "Remove tag "+t.name+"\n";
									}
								}
							}
							//metadati gps
							else if (i==size && gps!=null) {
								TiffOutputDirectory gpsDirectory = outputSet.getGPSDirectory();
								if (gpsDirectory != null) {
									gpsDirectory.removeField(TiffConstants.GPS_TAG_GPS_LATITUDE);
									gpsDirectory.removeField(TiffConstants.GPS_TAG_GPS_LONGITUDE);
									result = result + "Remove Tag GPS\n";
								}
								else
									result = result + "Can not remove tag GPS. Directory GPS not found.\n";
							}
						}
						catch (Exception e) {
							result = result + "Can not remove tag "+t.name+". ERROR: "+e.getMessage()+"\n";
						} 
					}
				}
				
				//thumbnail
				if (thumbChange) {
					try {
						ByteArrayOutputStream baos = new ByteArrayOutputStream();
						ImageIO.write(thumbnail, "jpg", baos );
						thumb = baos.toByteArray();
						TiffOutputDirectory dirThumb = outputSet.findDirectory(TiffDirectoryConstants.DIRECTORY_TYPE_SUB);
						if (dirThumb == null) {
							dirThumb = new TiffOutputDirectory(TiffDirectoryConstants.DIRECTORY_TYPE_SUB);
							outputSet.addDirectory(dirThumb);
						}
						if (delThumb) {
							dirThumb.setJpegImageData(null);
							result += "Delete Thumbnail\n";
						}
						else {
							dirThumb.setJpegImageData(new JpegImageData(offset, thumb.length, thumb));
							result += "Update Thumbnail\n";
						}
					}
					catch (Exception e) {
						result += "Can not update Thumbnail. ERROR: "+e.getMessage()+"\n";
					}
				}
			}
			
			//Salvataggio finale
			os = new FileOutputStream(out);
			os = new BufferedOutputStream(os);
			if(removeExif)
				new ExifRewriter().removeExifMetadata(in, os);
			else
				new ExifRewriter().updateExifMetadataLossless(in, os, outputSet);
			os.close();
			os = null;
			IJ.showMessage("Save", out+ " successfully saved!");
		}
		catch (Exception e) {
			result += "Can not save correctly the image. ERROR: "+e.getMessage()+"\n";
			IJ.showMessage("SAVE ERROR",e.toString()+" : "+e.getMessage());
		}
		finally {
			//Creazione report
			BufferedWriter report = null;
			try {
				report = new BufferedWriter(new FileWriter(out+" - Report.txt"));
				report.write(result);
				report.close();
			}
			catch (Exception e) {
				IJ.showMessage("ERROR","Can not write report!");
			}
			
			if (os != null) {
				try {
					os.close();
				}
				catch (IOException e) {
					IJ.showMessage(e.getMessage());
				}
			}
		}
	}
	
	public void insert() {
		GenericDialog gd = new GenericDialog("Insert new Tag");
		String[] stringTag;
		int size=tag.size();
		if (gps == null) {
			size++;
			stringTag = new String[size];
			stringTag[size-1] = "GPS";
		}
		else
			stringTag = new String[size];
		for (int i=0; i<tag.size(); i++)
			stringTag[i] = tag.get(i).name;
		
		gd.addChoice("Name", stringTag, stringTag[0]);
		gd.addMessage("Confirm insert tag?\nPress Ok to continue or Cancel to exit.");
		gd.showDialog();
		if (gd.wasCanceled())
			return;
		if (gd.wasOKed()) {
			int n = gd.getNextChoiceIndex();
			Object[] row = new Object[4];
			//gps
			if (n==size-1 && gps==null) {
				row[0] = "GPS Description";
				row[1] = "[GPS]";
				row[2] = new Boolean(true);
				row[3] = new Boolean(false);
				model.addRow(row);
				row[0] = "GPS Latitude (Degrees North)";
				row[1] = "(type double e.g. 1.5)";
				row[2] = new Boolean(false);
				row[3] = new Boolean(false);
				model.addRow(row);
				row[0] = "GPS Longitude (Degrees East)";
				row[1] = "(type double e.g. 1.5)";
				row[2] = new Boolean(false);
				row[3] = new Boolean(false);
				model.addRow(row);
			}
			else {
				TagInfo t = tag.get(n);
				tag.remove(n);
				int type = 0;
				if (t.dataTypes != null) type = t.dataTypes[0].type;
				row[0] = t;
				
				if (t.isText() || type == 2)	//String
					row[1] = "(type String e.g. text)";
				else if (t.length != 1 || type == 7)	//byte[]
					row[1] = "(type byte[] e.g. 1, 2, 3)";
				else if (type == 4 || type == 9)	//Long
					row[1] = "(type Long e.g. 1)";
				else if (type == 3 || type == 8)	//Short
					row[1] = "(type Short e.g. 1)";
				else if (type == 1 || type == 6)	//Byte
					row[1] = "(type Byte e.g. 1)";
				else if (type == 12)			//double
					row[1] = "(type Double e.g. 1.5)";
				else if (type == 11)			//float
					row[1] = "(type Float e.g. 1.5)";
				else if (type == 5 || type == 10)	//RationalNumber
					row[1] = "(type RationalNumber e.g. 1,5)";
				else
					row[1] = "";
				row[2] = new Boolean(false);
				row[3] = new Boolean(false);
				model.addRow(row);
			}
		}
	}
}

class ExifModel extends DefaultTableModel{

	static String[] labels = {"Name","Value","Set","Delete"};
	static Object[][] content = {};
	
	public ExifModel(){
		super(content,labels);
	 }
	
	public boolean isCellEditable(int row,int col){
		if (col == 0) return false;
		Object o = getItem(row);
		//tag esistenti: non selezionabili tag offset o sconosciuti
		if (o instanceof TiffImageMetadata.Item) {
			TiffImageMetadata.Item item = (TiffImageMetadata.Item) o;
			if (item.getTiffField().tagInfo.isOffset() || item.getTiffField().tagInfo.isUnknown()) {
				IJ.showMessage("Can not edit this tag");
				return false;
			}
			return true;
		}
		//tag nuovi: non selezionabile delete
		if (o instanceof TagInfo) {
			if (col == 3) {
				IJ.showMessage("This tag is not present in this image");
				return false;
			}
			return true;
		}
		//gps: non editabile info generale e non selezionabili set e delete di longitudine e latitudine
		if (o instanceof String) {
			String s = (String) o;
			if (s.equals("GPS Description")) {
				if (col == 1) {
					IJ.showMessage("Edit Latitude and Longitude");
					return false;
				}
				return true;
			}
			if (col == 1) return true;
			IJ.showMessage("Select \"GPS Description\"");
			return false;
		}
		return false;
	}
	public Class getColumnClass(int c) {
            return getValueAt(0, c).getClass();
        }
	public void setValueAt(Object aValue, int row, int column) {
		if (column == 1) {
			if (getItem(row) instanceof String) {}
			else setValueAt(new Boolean(true), row, 2);
		}
		else if (column == 2 && ((Boolean) getValueAt(row,3)==true))
			setValueAt(new Boolean(false), row, 3);
		else if (column == 3 && ((Boolean) getValueAt(row,2)==true))
			setValueAt(new Boolean(false), row, 2);
		super.setValueAt(aValue, row, column);
	}
	public Object getValueAt(int row, int column) {
		Object o = super.getValueAt(row,column);
		if (o instanceof TiffImageMetadata.Item) {
			TiffImageMetadata.Item item = (TiffImageMetadata.Item) o;
			return item.getTiffField().tagInfo.name;
		}
		if (o instanceof TagInfo) {
			TagInfo t = (TagInfo) o;
			return t.name;
		}
		return o;
	}
	public Object getItem(int row) {
		return super.getValueAt(row,0);
	}
}